#define STANDART_FILE_CALL 0
#define WORDS_BIGENDIAN 1

#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#define my64_t long long
#include "../includes/structures.h"

extern int transferoption ;
extern int filesize ;

int storemkdir(char *dirname,char *logmessage,int recursif) ;
int storecreatefile_rfio(char *filename,char *logmessage) ;

/*******************************************************************************
** storecreatefile :                                                           *
**                                                                             *
**      Routine to create a file                                               *
**                                                                             *
**      INPUT variable :                                                       *
**          filename    :  file to create    NOT MODIFIED                      *
**                                                                             *
**      OUTPUT variable :                                                      *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                                                     *
**          filesize                                                           *
**                                                                             *
**      TO FREE before any return : filepath                                   *
**                                                                             *
**      RETURN:                                                                *
**          -1  Creation failed unrecoverable error                            *
**           0  OK                                                             *
**           1  Creation failed recoverable error                              *
**                                                                             *
*******************************************************************************/
 
int storecreatefile(char *filename, char *logmessage) 
{
    char    *filepath ;
    int     fd ;
    int     lastslash ;
    int     savederrno ;
    int     retcode ;
#ifdef STANDART_FILE_CALL
 	off_t		toseek ;
	struct stat statbuf ;
#else
	off64_t		toseek ;
	struct stat64 statbuf ;
#endif
   
    /*
    ** Check if it is a rfio creation
    */
    if ( (transferoption & TROPT_RFIO) == TROPT_RFIO ) {
        return storecreatefile_rfio(filename,logmessage) ;
    }
    /*
    ** make a copy of filename for security
    */
    if ( (filepath = (char *) malloc (strlen(filename)+1) ) == NULL ) {
        sprintf(logmessage,"Error allocating memory for filepath : %s",strerror(errno)) ;
        return 1 ;
    }
    strcpy (filepath, filename);
    /*
    ** Check if the file exist
    **      We are going to first see if the directory exist
    **      and then look at the options 
    **      and then look if the file exist
    */
    lastslash = strlen(filepath) - 1 ;
    while ( lastslash >= 0 && filepath[lastslash] != '/') lastslash-- ;
    if ( lastslash == -1 ) {
        /*
        ** No slash in the path, that means that we have done a chdir
        ** before so do not care for the directory problem
        */
    } else if ( lastslash == 0 ) {
        /*
        ** A slash in first position so we suppose the "/" directory 
        ** exist ... nothing to do
        */
    } else if ( lastslash == strlen(filepath) - 1 ) {
        /*
        ** The filename end with a slash ..... error
        */
		syslog(LOG_ERR,"Filename %s ends with a /",filepath) ;
		sprintf(logmessage,"Filename %s ends with a /",filepath) ;
        free(filepath) ;
        return -1 ;
    } else {
        filepath[lastslash] = '\0';
        /*
        ** Check the existence of the directory 
        */
#ifdef STANDART_FILE_CALL
	    if ( stat(filepath,&statbuf) < 0 ) {
#else
	    if ( stat64(filepath,&statbuf) < 0 ) {
#endif
		    /*
		    ** It may be normal to get an error if the directory
		    ** does not exist but some error code must lead
		    ** to the interruption of the transfer:
		    **		EACCES		: Search permission denied
		    **		ELOOP		: To many symbolic links on path
		    **		ENAMETOOLONG: Path argument too long
		    **		ENOTDIR		: A component in path is not a directory
		    */
 		    savederrno = errno ;
		    if ( savederrno == EACCES ||
			    savederrno == ELOOP ||
			    savederrno == ENAMETOOLONG ||
			    savederrno == ENOTDIR ) {
			    syslog(LOG_ERR,"Error stating directory %s : %s ",filepath,strerror(savederrno)) ;
			    sprintf(logmessage,"Error stating directory %s : %s ",filepath,strerror(savederrno)) ;
                free(filepath) ;
                return -1 ;
            } else if (savederrno == ENOENT) {
                /*
                ** The directory does not exist so check for the TROPT_DIR
                ** option 
                */
                if ( (transferoption & TROPT_DIR ) != TROPT_DIR ) {
			        syslog(LOG_ERR,"Directory (%s) creation needed but TROPT_DIR not set",filepath) ;
			        sprintf(logmessage,"Directory (%s) creation needed but TROPT_DIR not set",filepath) ;
                    free(filepath) ;
                    return -1 ;
                } else {
                    if ( retcode = storemkdir(filepath,logmessage,1) != 0 ) return retcode ;
                    filepath[lastslash] = '/';
                }
            } else {
			    syslog(LOG_ERR,"Error stating directory %s : %s ",filepath,strerror(savederrno)) ;
			    sprintf(logmessage,"Error stating directory %s : %s ",filepath,strerror(savederrno)) ;
                free(filepath) ;
                return 1 ;
            }
        } else {
            /*
            ** The directory exist, check if it is a directory
            */
		    if ( (statbuf.st_mode & S_IFDIR) == S_IFDIR) {
                /*
                ** OK correct
                */
                filepath[lastslash] = '/';
            } else {
			    syslog(LOG_ERR,"%s is a not a directory ",filepath) ;
			    sprintf(logmessage,"%s is a not a directory",filepath) ;
                free(filepath) ;
                return -1 ;
            }
        }
    }
    /*
    ** At this stage all directory exists so check for the file 
    */
#ifdef STANDART_FILE_CALL
	if ( stat(filepath,&statbuf) < 0 ) {
#else
	if ( stat64(filepath,&statbuf) < 0 ) {
#endif
		/*
		** It may be normal to get an error if the file
		** does not exist but some error code must lead
		** to the interruption of the transfer:
		**		EACCES		: Search permission denied
		**		ELOOP		: To many symbolic links on path
		**		ENAMETOOLONG: Path argument too long
		**		ENOTDIR		: A component in path is not a directory
		*/
		savederrno = errno ;
		if ( savederrno == EACCES ||
			savederrno == ELOOP ||
			savederrno == ENAMETOOLONG ||
			savederrno == ENOTDIR ) {
			sprintf(logmessage,"Error stating file %s : %s ",filepath,strerror(savederrno)) ;
			syslog(LOG_ERR,"Error stating file %s : %s ",filepath,strerror(savederrno)) ;
            free(filepath) ;
			return -1 ;
		} else if (savederrno == ENOENT) {
			/*
			** That is normal the file does not exist
			*/
		} else {
			sprintf(logmessage,"Error stating file %s : %s ",filepath,strerror(savederrno)) ;
			syslog(LOG_ERR,"Error stating file %s : %s ",filepath,strerror(savederrno)) ;
            free(filepath) ;
			return 1 ;
		}
	} else {
		/*
		** The file exists so check if it is a directory
		*/
		if ( (statbuf.st_mode & S_IFDIR) == S_IFDIR) {
			syslog(LOG_ERR,"file %s is a directory",filepath) ;
			sprintf(logmessage,"File %s is a directory",filepath) ;
            free(filepath) ;
			return -1 ;
		}
	}				
	/*
	** We create the file
	*/
#ifdef STANDART_FILE_CALL
	if ((fd = open(filepath,O_WRONLY|O_CREAT|O_TRUNC,0777)) < 0 ) {
#else
	if ((fd = open64(filepath,O_WRONLY|O_CREAT|O_TRUNC,0777)) < 0 ) {
#endif
		/*
		** Depending on errno we are going to tell the client to 
		** retry or not
		*/
		savederrno = errno ;
		sprintf(logmessage,"Error creation file %s : %s ",filepath,strerror(errno)) ;
		syslog(LOG_ERR,"Error creation file %s : %s ",filepath,strerror(errno)) ;
		/*
		** We tell the client not to retry in the following case (even in waiting
		** WAITRETRYTIME the problem will not be solved) :
		**		EACCES		: Search permission denied
		**		EDQUOT		: No more quota 
		**		ENOSPC		: No more space
		**		ELOOP		: To many symbolic links on path
		**		ENAMETOOLONG: Path argument too long
		**		ENOTDIR		: A component in path is not a directory
		*/
		if ( savederrno == EACCES ||
				savederrno == EDQUOT ||
				savederrno == ENOSPC ||
				savederrno == ELOOP ||
				savederrno == ENAMETOOLONG ||
				savederrno == ENOTDIR ) {
                free(filepath) ;
			    return -1 ;
		} else {
            free(filepath) ;
			return 1 ;
		}
	}
    if ( filesize == 0 ) {
        close(fd) ;
        free(filepath) ;
        return 0 ;
    }
	/*
	** Lseek to set it to the correct size
    ** We use toseek because filesize is of type my64t and 
    ** call to lseek can be done with off_t which may be
    ** of length 64 bits or 32 bits
	*/
	toseek = filesize-1 ;
#ifdef STANDART_FILE_CALL
	if ( lseek(fd,toseek,SEEK_SET) < 0 ) {
#else
	if ( lseek64(fd,toseek,SEEK_SET) < 0 ) {
#endif
		sprintf(logmessage,"Error seeking file %s : %s ",filepath,strerror(errno)) ;
		syslog(LOG_ERR,"Error seeking file %s : %s ",filepath,strerror(errno)) ;
		close(fd) ;
		unlink(filepath) ;
        free(filepath) ;
		return 1 ;
	}
	/*
	** Write one byte 
	*/
	if ( write(fd,"\0",1) != 1) {
		savederrno = errno ;
		sprintf(logmessage,"Error writing file %s : %s ",filepath,strerror(errno)) ;
		syslog(LOG_ERR,"Error writing file %s : %s ",filepath,strerror(errno)) ;
		close(fd) ;
		unlink(filepath) ;
		/*
		** We tell the client not to retry in the following case (even in waiting
		** WAITRETRYTIME the problem will not be solved) :
		**		EDQUOT		: No more quota 
		**		ENOSPC		: No space on device
		*/
		if ( savederrno == EDQUOT ||
				savederrno == ENOSPC ) {
            free(filepath) ;
            return -1 ;
		} else {
            free(filepath) ;
            return 1 ;
		}
	}
	/*
	** And close the file
	*/
	if ( close(fd) < 0 ) {
		savederrno = errno ;
		unlink(filepath) ;
		sprintf(logmessage,"Error closing file %s : %s ",filepath,strerror(savederrno)) ;
		syslog(LOG_ERR,"Error closing file %s : %s ",filepath,strerror(savederrno)) ;
		if ( savederrno == ENOSPC ) {
            free(filepath) ;
            return -1 ;
		} else {
            free(filepath) ;
            return 1 ;
		}
	}
    return 0 ;
}
