#define STANDART_FILE_CALL 0

#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>

/*******************************************************************************
** storemkdir :                                                                *
**                                                                             *
**      Routine to create a bunch of directory                                 *
**                                                                             *
**      INPUT variable :                                                       *
**          dirname    :  directory to create    NOT MODIFIED                  *
**          recursif   :  1 recursif             NOT MODIFIED                  *
**                        0 non recursif                                       *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      TO FREE before any return : dirpath                                    *
**                                                                             *
**      RETURN:                                                                *
**          -1   Creation failed unrecoverable error                           *
**           0   OK                                                            *
**           >0  Creation failed recoverable error                             *
**                                                                             *
*******************************************************************************/
int storemkdir(char *dirname,char *logmessage,int recursif)
{
    char    *dirpath;
    char    *basedir ;
    char    *slash ;
    int     savederrno ;
#ifdef STANDART_FILE_CALL
    struct stat statbuf;
#else
	struct stat64 statbuf ;
#endif

    /*
    ** make a copy of dirname for security
    */
    if ( (dirpath = (char *) malloc (strlen(dirname)+1) ) == NULL ) {
        sprintf(logmessage,"Error allocating memory for dirpath : %s",strerror(errno)) ;
        return 1 ;
    }
    strcpy (dirpath, dirname);
    /*
    ** Strip trailing slash
    */
    strip_trailing_slashes(dirpath) ;
    slash = dirpath ;
    if ( recursif == 1 ) {
        /*
        ** Recursif, we are going to create all directory
        */
        /*
        ** If this is an absolute path strip the leading slash
        */
        if ( *dirpath == '/' ) {
            while (*slash == '/') slash++;
        }
        while (1) {
	        slash = strchr (slash, '/');
            if ( slash == NULL ) break ;
            basedir = dirpath ;
	        /*
            ** The mkdir and stat calls below appear to be reversed.
	        ** They are not.  It is important to call mkdir first and then to
	        ** call stat (to distinguish the three cases) only if mkdir fails.
	        ** The alternative to this approach is to `stat' each directory,
	        ** then to call mkdir if it doesn't exist.  But if some other process
	        ** were to create the directory between the stat & mkdir, the mkdir
	        ** would fail with EEXIST. 
            ** We create it with mode = 777 because the umask is set by bbftpd
            */
 	        *slash = '\0';
	        if ( mkdir (basedir, 0777) ) {
                savederrno = errno ;
#ifdef STANDART_FILE_CALL
                if ( stat (basedir,&statbuf ) ) {
#else
                if ( stat64 (basedir,&statbuf ) ) {
#endif
                    sprintf(logmessage,"Error creation directory %s (%s)",basedir,strerror(savederrno)) ; 
		            /*
		            ** We tell the client not to retry in the following case (even in waiting
		            ** WAITRETRYTIME the problem will not be solved) :
		            **		EACCES		: Search permission denied
		            **		EDQUOT		: No more quota 
		            **		ENOSPC		: No more space
		            **		ELOOP		: To many symbolic links on path
		            **		ENAMETOOLONG: Path argument too long
		            **		ENOTDIR		: A component in path is not a directory
		            **		EROFS		: The path prefix resides on a read-only file system.
		            **		ENOENT      : A component of the path prefix does not exist or is a null pathname.
		            **		EEXIST      : The named file already exists.
		            */
 		            if ( savederrno == EACCES ||
				            savederrno == EDQUOT ||
				            savederrno == ENOSPC ||
				            savederrno == ELOOP ||
				            savederrno == ENAMETOOLONG ||
				            savederrno == ENOTDIR ||
				            savederrno == EROFS ||
				            savederrno == ENOENT ||
				            savederrno == EEXIST ) {
                        free(dirpath) ;
                        return -1 ;
                    } else {
                        free(dirpath) ;
                        return 1 ;
                    }
                } else if ( (statbuf.st_mode & S_IFDIR) != S_IFDIR) {
                    sprintf(logmessage,"%s exists but is not a directory",basedir) ; 
                    free(dirpath) ;
                    return -1 ;
                } else {
 		            /*
                    ** dirpath already exists and is a directory. 
                    */
                }
		    }
 	        *slash++ = '/';

            /* 
            ** Avoid unnecessary calls to `stat' when given
	        ** pathnames containing multiple adjacent slashes.  
            */
	        while (*slash == '/') slash++;
       }
    }
    /*
    ** We have created all leading directories so let see the last one
    */
    basedir = dirpath ;
	if ( mkdir (basedir, 0777) ) {
        savederrno = errno ;
#ifdef STANDART_FILE_CALL
        if ( stat (basedir,&statbuf ) ) {
#else
        if ( stat64 (basedir,&statbuf ) ) {
#endif
            sprintf(logmessage,"Error creation directory %s (%s)",basedir,strerror(savederrno)) ; 
		    /*
		    ** We tell the client not to retry in the following case (even in waiting
		    ** WAITRETRYTIME the problem will not be solved) :
		    **		EACCES		: Search permission denied
		    **		EDQUOT		: No more quota 
		    **		ENOSPC		: No more space
		    **		ELOOP		: To many symbolic links on path
		    **		ENAMETOOLONG: Path argument too long
		    **		ENOTDIR		: A component in path is not a directory
		    **		EROFS		: The path prefix resides on a read-only file system.
		    **		ENOENT      : A component of the path prefix does not exist or is a null pathname.
		    **		EEXIST      : The named file already exists.
		    */
 		    if ( savederrno == EACCES ||
				    savederrno == EDQUOT ||
				    savederrno == ENOSPC ||
				    savederrno == ELOOP ||
				    savederrno == ENAMETOOLONG ||
				    savederrno == ENOTDIR ||
				    savederrno == EROFS ||
				    savederrno == ENOENT ||
				    savederrno == EEXIST ) {
                free(dirpath) ;
                return -1 ;
            } else {
                free(dirpath) ;
                return 1 ;
            }
        } else if ( (statbuf.st_mode & S_IFDIR) != S_IFDIR) {
            sprintf(logmessage,"%s exists but is not a directory",basedir) ; 
            free(dirpath) ;
            return -1 ;
        } else {
 		    /*
            ** dirpath already exists and is a directory. 
            */
            free(dirpath) ;
            return 0 ;
        }
	}
    free(dirpath) ;
    return 0 ;
}
